package de.tud.kom.p2psim.impl.network.fairshare;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.impl.util.logging.SimLogger;
import de.tudarmstadt.maki.simonstrator.api.Time;

/**
 * An implementation of the Connection interface. TODO to be extended.
 * 
 * @author Tobias Lauinger
 */
public class Link implements Connection {

	static Logger log = SimLogger.getLogger(Link.class);

	private final Node src;

	private final Node dst;

	private long lastTime;

	private long transferEndTime;

	private double bandwidth;

	private double oldBandwidth;

	private double remainingBytes;

	// private double transferedBytes;

	private double messageSize;

	private long creationTime;

	private boolean burstMessage;

	private boolean statusChanged;

	private final long propagationDelay;

	public Link(final Node source, final Node sink, double msgSize,
			final long minimumDelay) {
		src = source;
		dst = sink;
		src.addConnection(this, false);
		dst.addConnection(this, true);
		creationTime = Time.getCurrentTime();
		remainingBytes = msgSize;
		messageSize = msgSize;
		lastTime = creationTime;
		burstMessage = false;
		statusChanged = false;
		propagationDelay = minimumDelay;
	}

	public long getCreationTime() {
		return creationTime;
	}

	public long getPropagationDelay() {
		return propagationDelay;
	}

	public void addBurstMessage(double msgSize) {
		creationTime = Time.getCurrentTime();
		remainingBytes = msgSize;
		assert (remainingBytes > 0);
		messageSize = msgSize;
		lastTime = creationTime;
		burstMessage = true;
		statusChanged = true;
	}

	@Override
	public double getBandwidth() {
		return bandwidth;
	}

	public void calculateNewTransferEndTime() {
		long currentTime = Time.getCurrentTime();
		long lastTransferInterval = currentTime - lastTime;
		lastTime = currentTime;
		assert (lastTransferInterval >= 0);

		double byteBurst = (lastTransferInterval / (double) Time.SECOND)
				* oldBandwidth;

		remainingBytes -= byteBurst;

		double time = remainingBytes / bandwidth;

		transferEndTime = Math.round(time * Time.SECOND) + currentTime;
		if (transferEndTime < currentTime)
			transferEndTime = currentTime;

		// log.info("**************************************************************");
		// log.info("Link " + src.getNetID() + " -> " + dst.getNetID() + " @ "
		// + Simulator.getCurrentTime());
		// log.info("lastTransferInterval (ns): " + lastTransferInterval
		// + " |  Total bytes: " + messageSize + " | Remaining bytes "
		// + remainingBytes);
		// log.info("Allocated link bandwidth: " + bandwidth
		// + " bytes/s | Transmission duration: " + time
		// * Simulator.MILLISECOND_UNIT + " ms");
		// log.info("Transfer endtime: " + transferEndTime);
		// log.info("**************************************************************");
		// log.info("");

		statusChanged = false;
		assert (creationTime < transferEndTime);
	}

	public long getTransferEndTime() {
		return this.transferEndTime;
	}

	@Override
	public double getMinBid() {
		final double bidSrc = src.getCurrentFairShare(false);
		final double bidDst = dst.getCurrentFairShare(true);
		assert (bidSrc > 0);
		assert (bidDst > 0);
		if (bidSrc > bidDst) {
			return bidDst;
		} else {
			return bidSrc;
		}
	}

	@Override
	public Node getNode(boolean destination) {
		if (destination) {
			return dst;
		} else {
			return src;
		}
	}

	@Override
	public boolean isInUse() {
		return true;
	}

	public boolean isBurst() {
		return burstMessage;
	}

	public void resetBurst() {
		burstMessage = false;
	}

	@Override
	public void reset() {
		oldBandwidth = bandwidth;
		statusChanged = true;
		bandwidth = 0;
	}

	public boolean hasStatusChanged() {
		return statusChanged;
	}

	@Override
	public int compareTo(Connection c) {
		if (c instanceof Link)
			return ((Double) getMinBid()).compareTo((Double) c.getMinBid());

		log.error("Cannot compare Link to " + c.getClass());
		return 0;
	}

	@Override
	public void assignBandwidth(final double newBandwidth) {

		assert (newBandwidth > 0);
		bandwidth = newBandwidth;
		src.updateAvailableBandwidth(bandwidth, false);
		dst.updateAvailableBandwidth(bandwidth, true);

		log.debug("Setting rate to " + newBandwidth + "bytes/sec for " + this);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Link) {
			Link aLink = (Link) obj;
			if (this.getNode(true).equals(aLink.getNode(true))) {
				if (this.getNode(false).equals(aLink.getNode(false))) {
					return true;
				}
			}
		}
		return false;
	}

	@Override
	public String toString() {
		return "Flow[ " + this.src.getNetID() + "(U:"
				+ this.src.getCurrentBandwidth().getUpBW() + "/D:"
				+ this.src.getCurrentBandwidth().getDownBW() + ")" + "\t"
				+ " -> " + this.dst.getNetID() + "(U:"
				+ this.dst.getCurrentBandwidth().getUpBW() + "/D:"
				+ this.dst.getCurrentBandwidth().getDownBW() + ")" + "\t"
				+ " ]";
	}

}
