package de.tud.kom.p2psim.impl.network.fairshare;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.analyzer.MessageAnalyzer.Reason;
import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.api.network.BandwidthImpl;
import de.tud.kom.p2psim.api.network.FlowBasedNetlayer;
import de.tud.kom.p2psim.api.network.NetMessage;
import de.tud.kom.p2psim.api.network.NetProtocol;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.network.AbstractNetLayer;
import de.tud.kom.p2psim.impl.network.IPv4Message;
import de.tud.kom.p2psim.impl.network.IPv4NetID;
import de.tud.kom.p2psim.impl.network.fairshare.position.FairshareGnpPosition;
import de.tud.kom.p2psim.impl.network.fairshare.position.GeoLocation;
import de.tud.kom.p2psim.impl.transport.AbstractTransMessage;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;
import de.tudarmstadt.maki.simonstrator.api.Message;
import de.tudarmstadt.maki.simonstrator.api.component.network.NetID;

/**
 * A Node is a participant in a network. It has connections to other Nodes, but
 * never to itself. Its incoming and outgoing network bandwidth is restricted.
 * It can participate in an algorithm to assign each connection a fair share of
 * bandwidth.
 * 
 * @author Tobias Lauinger, Sebastian Kaune, Yue Sheng
 */
public class Node extends AbstractNetLayer implements FlowBasedNetlayer {

	private static Logger log = SimLogger.getLogger(Node.class);
	
	private final HashSet<Connection> incomingConnections;

	private final HashSet<Connection> outgoingConnections;

	private final OptimizedSubnet subnet;

	private double unassignedIncomingConnections;

	private double unassignedOutgoingConnections;

	private final GeoLocation geoLocation;

	private final Map<Node, List<NetMessage>> hostQueues;

	public Node(SimHost host, final double maxDownBW, final double maxUpBW,
			IPv4NetID netID,
			FairshareGnpPosition netPosition, GeoLocation geoLoc,
			OptimizedSubnet gnpSubNet) {
		super(host, netID, new BandwidthImpl(maxDownBW, maxUpBW), netPosition,
				null);
		subnet = gnpSubNet;
		incomingConnections = new LinkedHashSet<Connection>();
		outgoingConnections = new LinkedHashSet<Connection>();
		geoLocation = geoLoc;
		hostQueues = new HashMap<Node, List<NetMessage>>();
		subnet.registerNetLayer(this);
	}

	public void setIncomingBandwidth(double incomingBw) {
		getMaxBandwidth().setDownBW(incomingBw);
	}

	public void setOutgoingBandwidth(double outgoingBw) {
		getMaxBandwidth().setUpBW(outgoingBw);
	}

	/**
	 * Adds the given Connection to the Connections of this Node. The new
	 * Connection does not automatically obtain bandwidth; this has to be
	 * triggered via a run of the fair share algorithm.
	 * 
	 * @param c
	 *            the Connection to be added.
	 * @param incoming
	 *            true if the Connection is a download connection.
	 */
	public void addConnection(Connection c, boolean incoming) {
		if (incoming) {
			incomingConnections.add(c);
			// unassignedIncomingConnections++;
		} else {
			outgoingConnections.add(c);
			// unassignedOutgoingConnections++;
		}
	}

	/**
	 * Removes the given Connection from the Connections of this Node.
	 * 
	 * @param c
	 *            the Connection to be removed.
	 * @param incoming
	 *            true if the Connection is a download connection.
	 */
	public void removeConnection(Connection c, boolean incoming) {
		if (incoming) {
			incomingConnections.remove(c);
		} else {
			outgoingConnections.remove(c);
		}
	}

	/**
	 * Adds the Connections of this Node, either incoming or outgoing, to the
	 * given Collection and calls this method on the Nodes on the other sides of
	 * the Connections, if they were not previously contained in the Collection.
	 * Also resets the state of the involved Connections and Nodes.
	 * 
	 * @param c
	 *            the Collection with the Connections that have already been
	 *            visited.
	 * @param incoming
	 *            true if download Connections are to be followed (which means
	 *            that at the other end of this Node's Connections upload
	 *            Connections will be followed).
	 */
	public void traverseConnections(Collection<Connection> c, boolean incoming) {
		final Collection<Connection> toTraverse;
		if (incoming) {
			toTraverse = incomingConnections;
		} else {
			toTraverse = outgoingConnections;
		}
		this.reset(incoming);
		for (final Connection con : toTraverse) {
			// add only if in use, traverse only if added
			if (con.isInUse() && c.add(con)) {
				// reset state of Connection
				con.reset(); // TODO will be called twice

				// traverse other end of connection
				// traverse other type of connections at other end
				con.getNode(!incoming).traverseConnections(c, !incoming);
			}
		}
	}

	/**
	 * Resets the state of this Node for a new, clean run of the fair share
	 * algorithm. The state that will be cleared are the number of assigned
	 * connections and the currently unassigned incoming and outgoing network
	 * capacity.
	 * 
	 * @param incoming
	 *            true if the state for the incoming (download) Connections is
	 *            to be reset (else the outgoing/upload connections).
	 */
	public void reset(boolean incoming) {
		if (incoming) {
			getCurrentBandwidth().setDownBW(getMaxBandwidth().getDownBW());
			unassignedIncomingConnections = incomingConnections.size();
		} else {
			getCurrentBandwidth().setUpBW(getMaxBandwidth().getUpBW());
			unassignedOutgoingConnections = outgoingConnections.size();
		}
	}

	/**
	 * Calculates the current fair share of bandwidth for the unassigned
	 * connections (the available bandwidth divided by the number of unassigned
	 * connections).
	 * 
	 * @param incoming
	 *            true if download connections are to be considered, else
	 *            considering upload connections.
	 * @return the current fair share for the unassigned connections.
	 */
	public double getCurrentFairShare(boolean incoming) {
		// TODO "incoming" not currently used
		if (incoming) {
			assert unassignedIncomingConnections > 0 : "unassigned download-connections:"
					+ unassignedIncomingConnections;
			return getCurrentBandwidth().getDownBW()
					/ unassignedIncomingConnections;
		} else {
			assert unassignedOutgoingConnections > 0;
			return getCurrentBandwidth().getUpBW()
					/ unassignedOutgoingConnections;
		}
	}

	/**
	 * Updates this Node with the newly assigned bandwidth. It will be
	 * subtracted from the currently unassigned bandwidth.
	 * 
	 * @param bandwidth
	 *            the bandwidth that has been assigned to the Connection.
	 * @param incoming
	 *            true if the updated Connection was a download connection.
	 */
	public void updateAvailableBandwidth(double bandwidth, boolean incoming) {
		double newBandwidth;
		if (incoming) {
			assert unassignedIncomingConnections > 0;
			// unassignedIncomingBandwidth -= bandwidth;
			newBandwidth = getCurrentBandwidth().getDownBW() - bandwidth;
			getCurrentBandwidth().setDownBW(newBandwidth);
			unassignedIncomingConnections--;
		} else {
			assert unassignedOutgoingConnections > 0;
			// unassignedOutgoingBandwidth -= bandwidth;
			newBandwidth = getCurrentBandwidth().getUpBW() - bandwidth;
			getCurrentBandwidth().setUpBW(newBandwidth);
			unassignedOutgoingConnections--;
		}
		assert newBandwidth >= 0;
	}

	@Override
	protected boolean isSupported(TransProtocol protocol) {
		return (protocol.equals(TransProtocol.UDP) || protocol
				.equals(TransProtocol.TCP));
	}

	/**
	 * 
	 * @return 2-digit country code
	 */
	public String getCountryCode() {
		return geoLocation.getCountryCode();
	}

	@Override
	public void send(Message msg, NetID receiverId, NetProtocol protocol) {
		if (isOnline()) {
			assert (msg.getSize() >= 0);
			assert (isSupported(((AbstractTransMessage) msg).getProtocol()));

			NetMessage netMsg = new IPv4Message(msg, receiverId,
					this.getLocalInetAddress());
			TransProtocol current = ((AbstractTransMessage) msg).getProtocol();
			if (current.equals(TransProtocol.UDP)) {
				if (hasAnalyzer) {
					netAnalyzerProxy
							.netMsgEvent(netMsg, getHost(), Reason.SEND);
				}
				this.subnet.sendUDP(netMsg);
			} else if (current.equals(TransProtocol.TCP)) {
				Node receiver = subnet.getNetLayer(receiverId);
				if (!hostQueues.containsKey(receiver))
					hostQueues.put(receiver, new LinkedList<NetMessage>());
				List<NetMessage> queuedMessages = hostQueues.get(receiver);
				if (queuedMessages.isEmpty()) {
					if (hasAnalyzer) {
						netAnalyzerProxy.netMsgEvent(netMsg, getHost(),
								Reason.SEND);
					}
					subnet.sendTCPMessage(this, receiver, netMsg.getSize());
				}
				queuedMessages.add(netMsg);
			}
		}
	}

	public boolean isMessageQueueEmpty(Node receiver) {
		return hostQueues.get(receiver).isEmpty();
	}

	public double peekMessageQueue(Node receiver) {
		return hostQueues.get(receiver).get(0).getSize();
	}

	public List<NetMessage> getViewOnMessageQueue(Node receiver) {
		return Collections.unmodifiableList(hostQueues.get(receiver));
	}

	public NetMessage removeMessageFromQueue(Node receiver) {
		return hostQueues.get(receiver).remove(0);
	}

	/**
	 * Returns the minimum end-to-end propagation delay in microseconds between
	 * this host, and the host the specified netID belongs to.
	 * 
	 * @param netId
	 *            the remote netID
	 * @return -1 if no information available, else the minimum delay (in
	 *         microseconds)
	 */
	public long getMinimumPropagationDelay(NetID netId) {
		if (subnet.getNetLayer(netId) != null)
			return subnet.getLatencyModel().getLatency(this,
					subnet.getNetLayer(netId));
		else
			return -1l;
	}

	@Override
	public void goOffline() {
		super.goOffline();
		hostQueues.clear();
		subnet.isOffline(this,
				(HashSet<Connection>) outgoingConnections.clone(),
				(HashSet<Connection>) incomingConnections.clone());
		assert (incomingConnections.size() == 0);
		assert (outgoingConnections.size() == 0);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Node) {
			Node aNode = (Node) obj;
			if (this.getLocalInetAddress().equals(aNode.getNetID())) {
				return true;
			}
		}
		return false;
	}

	@Override
	public int hashCode() {
		return getLocalInetAddress().hashCode();
	}
}
