package de.tud.kom.p2psim.impl.network.fairshare;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;

import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.impl.network.IPv4NetID;
import de.tud.kom.p2psim.impl.network.fairshare.position.FairshareGnpPosition;
import de.tud.kom.p2psim.impl.network.fairshare.position.GeoLocation;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.component.HostComponentFactory;

public class OptimizedNetLayerFactory implements HostComponentFactory {

	private static Logger log = SimLogger
			.getLogger(OptimizedNetLayerFactory.class);

	private final OptimizedSubnet subnet;

	private double downBandwidth;

	private double upBandwidth;

	private HashMap<IPv4NetID, GnpHostInfo> hostPool;

	private HashMap<String, ArrayList<IPv4NetID>> namedGroups;

	public OptimizedNetLayerFactory() {
		subnet = new OptimizedSubnet();
	}

	@Override
	public Node createComponent(Host phost) {
		SimHost host = (SimHost) phost;
		Node netLayer = newNetLayer(host, host.getProperties().getGroupID());
		return netLayer;
	}

	/**
	 * random node form group
	 * 
	 * @param id
	 * @return
	 */
	public Node newNetLayer(SimHost host, String id) {
		if (this.namedGroups.containsKey(id)
				&& !this.namedGroups.get(id).isEmpty()) {
			int size = namedGroups.get(id).size();
			IPv4NetID netId = namedGroups.get(id).get(
					Randoms.getRandom(OptimizedNetLayerFactory.class).nextInt(
							size));
			namedGroups.get(id).remove(netId);
			return newNetLayer(host, netId);
		} else {
			throw new IllegalStateException(
					"No (more) Hosts are assigned to \"" + id + "\"");
		}
	}

	private Node newNetLayer(SimHost host, IPv4NetID netID) {
		FairshareGnpPosition gnpPos = this.hostPool.get(netID).gnpPosition;
		GeoLocation geoLoc = this.hostPool.get(netID).geoLoc;
		Node nw = new Node(host, downBandwidth, upBandwidth, netID, gnpPos,
				geoLoc,
				subnet);
		hostPool.remove(netID);
		return nw;
	}

	public void setGnpFile(String gnpFileName) {

		File gnpFile = new File(gnpFileName);

		hostPool = new HashMap<IPv4NetID, GnpHostInfo>();
		namedGroups = new HashMap<String, ArrayList<IPv4NetID>>();

		log.info("Read hosts from file " + gnpFile);
		SAXReader reader = new SAXReader(false);

		Document configuration = null;
		try {
			configuration = reader.read(gnpFile);
		} catch (DocumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Element root = configuration.getRootElement();
		assert root.getName().equals("gnp");

		for (Object obj : root.elements()) {
			Element elem = (Element) obj;
			if (elem.getName().equals("GroupLookup")) {
				for (Iterator iter = elem.elementIterator("Group"); iter
						.hasNext();) {
					Element variable = (Element) iter.next();
					String id = variable.attributeValue("id");

					ArrayList<IPv4NetID> group = new ArrayList<IPv4NetID>();
					for (Iterator ipIter = variable.elementIterator("IPs"); ipIter
							.hasNext();) {
						Element ipElement = (Element) ipIter.next();
						String[] ips = ipElement.attributeValue("value").split(
								",");
						for (int c = 0; c < ips.length; c++)
							group.add(new IPv4NetID(Long.parseLong(ips[c])));
					}
					if (namedGroups.containsKey(id)) {
						throw new IllegalStateException(
								"Multiple Group Definition in " + gnpFileName
										+ " ( Group: " + id + " )");
					} else {
						namedGroups.put(id, group);
					}
				}
			} else if (elem.getName().equals("Hosts")) {
				for (Iterator iter = elem.elementIterator("Host"); iter
						.hasNext();) {
					Element variable = (Element) iter.next();

					// IP-Address
					IPv4NetID hostID = new IPv4NetID(Long.parseLong(variable
							.attributeValue("ip")));

					// GNP-Coordinates
					String[] coordinatesS = variable.attributeValue(
							"coordinates").split(",");
					double[] coordinatesD = new double[coordinatesS.length];
					for (int c = 0; c < coordinatesD.length; c++)
						coordinatesD[c] = Double.parseDouble(coordinatesS[c]);
					FairshareGnpPosition gnpPos = new FairshareGnpPosition(coordinatesD);

					// GeoLocation
					String continentalArea = variable
							.attributeValue("continentalArea");
					String countryCode = variable.attributeValue("countryCode");
					String region = variable.attributeValue("region");
					String city = variable.attributeValue("city");
					String isp = variable.attributeValue("isp");
					double longitude = Double.parseDouble(variable
							.attributeValue("longitude"));
					double latitude = Double.parseDouble(variable
							.attributeValue("latitude"));
					GeoLocation geoLoc = new GeoLocation(continentalArea,
							countryCode, region, city, isp, latitude, longitude);

					GnpHostInfo hostInfo = new GnpHostInfo(geoLoc, gnpPos);
					hostPool.put(hostID, hostInfo);
				}
			}
		}
	}

	public void setDownBandwidth(double downBandwidth) {
		this.downBandwidth = downBandwidth / 8.0;
	}

	public void setUpBandwidth(double upBandwidth) {
		this.upBandwidth = upBandwidth / 8.0;
	}

	private class GnpHostInfo {

		private final FairshareGnpPosition gnpPosition;

		private final GeoLocation geoLoc;

		public GnpHostInfo(GeoLocation geoLoc, FairshareGnpPosition gnpPos) {
			this.gnpPosition = gnpPos;
			this.geoLoc = geoLoc;
		}

	}

}
