package de.tud.kom.p2psim.impl.network.fairshare;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.network.NetLatencyModel;
import de.tud.kom.p2psim.api.network.NetLayer;
import de.tud.kom.p2psim.api.network.NetMessage;
import de.tud.kom.p2psim.impl.network.AbstractSubnet;
import de.tud.kom.p2psim.impl.network.IPv4Message;
import de.tud.kom.p2psim.impl.network.IPv4NetID;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;
import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.EventHandler;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.network.NetID;

/**
 * 
 * @author Gerald Klunker
 * @version 0.1, 17.01.2008
 * 
 */
public class OptimizedSubnet extends AbstractSubnet implements EventHandler {

	private static Logger log = SimLogger.getLogger(OptimizedSubnet.class);

	private final NetLatencyModel netLatencyModel;

	private final Map<IPv4NetID, Node> nodes;

	private final FairShareAlgorithm fairShare;

	private final List<Link> links;

	private long nextMsgArrival;

	private final static int EVENT_RECEIVE = 1;

	private final static int EVENT_STATUS = 2;

	public OptimizedSubnet() {
		nodes = new HashMap<IPv4NetID, Node>();
		fairShare = new FairShareAlgorithm();
		links = new LinkedList<Link>();
		netLatencyModel = new OptimizedLatencyModel();
	}

	public OptimizedSubnet(NetLatencyModel model) {
		nodes = new HashMap<IPv4NetID, Node>();
		fairShare = new FairShareAlgorithm();
		links = new LinkedList<Link>();
		netLatencyModel = model;
	}

	Node getNetLayer(NetID netId) {
		return nodes.get(netId);
	}

	public NetLatencyModel getLatencyModel() {
		return netLatencyModel;
	}

	/**
	 * Registers a NetWrapper in the SubNet.
	 * 
	 * @param wrapper
	 *            The NetWrapper.
	 */
	@Override
	public void registerNetLayer(NetLayer netLayer) {
		nodes.put((IPv4NetID) netLayer.getNetID(), (Node) netLayer);
	}

	@Override
	public void eventOccurred(Object content, int type) {
		if (type == EVENT_RECEIVE) {
			IPv4Message msg = (IPv4Message) content;
			Node receiver = nodes.get(msg.getReceiver());
			receiver.receive(msg);
		} else if (type == EVENT_STATUS) {
			if (Time.getCurrentTime() == nextMsgArrival) {
				log.debug("New Message arrival at -----------------------------------------------------> "
						+ Time.getCurrentTime());
				log.debug("Number of Links: " + links.size());

				List<Link> obsoleteLinks = new LinkedList<Link>();
				boolean error = true;

				// Sort Link list by means of earliest transmission end time
				Collections.sort(links, new TransferEndTimeComp());

				// Dispatch messages and filter obsolete Links
				for (Link link : links) {
					if (link.getTransferEndTime() > nextMsgArrival) {
						assert (!error);
						break;
					} else {
						error = false;
						assert (link.getTransferEndTime() == nextMsgArrival);
						Node sender = link.getNode(false);
						Node receiver = link.getNode(true);
						assert (sender.isMessageQueueEmpty(receiver) == false);

						NetMessage message = sender
								.removeMessageFromQueue(receiver);
						Event.scheduleWithDelay(link.getPropagationDelay(),
								this, message, EVENT_RECEIVE);

						assert (Time.getCurrentTime() - link.getCreationTime() > 0);
						log.info(Time.getCurrentTime()
								+ ">TCP Message arrived ("
								+ message.getSender()
								+ " -> "
								+ message.getReceiver()
								+ ") | "
								+ message.getPayload()
								+ ") | "
								+ message.getPayload().getPayload()
								+ " | Transmission time: "
								+ ((Time.getCurrentTime() - link
										.getCreationTime()) / (double) Time.MILLISECOND)
								+ " ms"
								+ "; Scheduling for "
								+ (Time.getCurrentTime() + link
										.getPropagationDelay()) + ".");

						link.resetBurst();

						boolean anotherMessageInQueue = !sender
								.isMessageQueueEmpty(receiver);
						if (anotherMessageInQueue) {
							link.addBurstMessage(sender
									.peekMessageQueue(receiver));
						} else {
							obsoleteLinks.add(link);
						}

					}
				}

				for (Link link : obsoleteLinks) {
					if (!link.isBurst()) {
						Node sender = link.getNode(false);
						Node receiver = link.getNode(true);
						links.remove(link);
						sender.removeConnection(link, false);
						receiver.removeConnection(link, true);

						/*
						 * NOTE: CHANGED by Peter Heise:
						 * 
						 * Did not recalculate all links on remove. Now has to
						 * called like.
						 */

						LinkedHashSet<Connection> links = new LinkedHashSet<Connection>();
						receiver.traverseConnections(links, true);
						sender.traverseConnections(links, false);
						fairShare.run(links);

					}
				}
				scheduleNextMessageArrival();
			}
		}
	}

	private void scheduleNextMessageArrival() {
		nextMsgArrival = Long.MAX_VALUE;
		for (Link l : links) {
			if (l.hasStatusChanged())
				l.calculateNewTransferEndTime();
			if (l.getTransferEndTime() < nextMsgArrival)
				nextMsgArrival = l.getTransferEndTime();
		}
		assert (nextMsgArrival >= Time.getCurrentTime()) : "nextMsgArrival is "
				+ nextMsgArrival
				+ "; current time is "
				+ Time.getCurrentTime();
		Event.scheduleWithDelay(nextMsgArrival - Time.getCurrentTime(), this,
				null, EVENT_STATUS);

	}

	@Override
	public void send(NetMessage msg) {
		// TODO Auto-generated method stub
	}

	public void sendUDP(NetMessage netMsg) {
		Node sender = nodes.get(netMsg.getSender());
		Node receiver = nodes.get(netMsg.getReceiver());
		long end2endDelay = netLatencyModel.getLatency(sender, receiver);
		Event.scheduleWithDelay(end2endDelay, this, netMsg, EVENT_RECEIVE);
	}

	private void printFairShareGraph(String header) {
		System.out.println(header);
		for (Link link : links) {
			System.out.println("Link: " + link.getNode(false).getNetID()
					+ " --> " + link.getNode(true).getNetID() + " ("
					+ link.getBandwidth() + ")");
		}
	}

	// public boolean isLinkAlreadyEstablished(Node sender, Node receiver) {
	// for (Link link : links) {
	// if (link.getNode(false).equals(sender) &&
	// link.getNode(true).equals(receiver))
	// return true;
	// }
	// return false;
	// }

	public void sendTCPMessage(Node sender, Node receiver, double messageSize) {
		
		
		Link link = new Link(sender, receiver, messageSize,
				netLatencyModel.getLatency(sender, receiver));
		// assert (!isLinkAlreadyEstablished(sender, receiver));
		assert (!links.contains(link));
		links.add(link);
		
		LinkedHashSet<Connection> links = new LinkedHashSet<Connection>();
		sender.traverseConnections(links, false);
		fairShare.run(links);
		
		log.debug("Sending TCP Message to " + receiver.getNetID()
				+ " with latency of "
				+ netLatencyModel.getLatency(sender, receiver)
				/ Time.MILLISECOND + "ms at " + Time.getCurrentTime()
				/ Time.MILLISECOND + "ms, size=" + messageSize + ", rate = "
				+ link.getBandwidth());
		
		scheduleNextMessageArrival();
	}

	public void isOffline(Node offlineNode,
			Set<Connection> outgoingConnections,
			Set<Connection> incomingConnections) {
		// int a = 1;

		for (Connection link : outgoingConnections) {
			Node otherSide = link.getNode(true);
			links.remove(link);
			offlineNode.removeConnection(link, false);
			otherSide.removeConnection(link, true);
			
			LinkedHashSet<Connection> links = new LinkedHashSet<Connection>();
			otherSide.traverseConnections(links, true);
			fairShare.run(links);
		}
		for (Connection link : incomingConnections) {
			Node otherSide = link.getNode(false);
			links.remove(link);
			offlineNode.removeConnection(link, true);
			
			LinkedHashSet<Connection> links = new LinkedHashSet<Connection>();
			otherSide.traverseConnections(links, false);
			fairShare.run(links);
		}
		scheduleNextMessageArrival();
	}

	public void cancelTransmission(Node sender, Node receiver) {
		Link toCancel = null;
		for (Link link : links) {
			Node src = link.getNode(false);
			Node dst = link.getNode(true);
			if (src.equals(sender) && dst.equals(receiver)) {
				toCancel = link;
				break;
			}
		}

		assert (toCancel != null);

		NetMessage message = sender.removeMessageFromQueue(receiver);
		toCancel.resetBurst();

		boolean anotherMessageInQueue = !sender.isMessageQueueEmpty(receiver);
		if (anotherMessageInQueue) {
			toCancel.addBurstMessage(sender.peekMessageQueue(receiver));
		} else {
			links.remove(toCancel);
			sender.removeConnection(toCancel, false);
			receiver.removeConnection(toCancel, true);
			
			LinkedHashSet<Connection> links = new LinkedHashSet<Connection>();
			sender.traverseConnections(links, false);
			fairShare.run(links);
			

		}
		scheduleNextMessageArrival();
	}

	class TransferEndTimeComp implements Comparator<Link> {
		@Override
		public int compare(Link o1, Link o2) {
			if (o1.getTransferEndTime() < o2.getTransferEndTime()) {
				return -1;
			} else if (o1.getTransferEndTime() > o2.getTransferEndTime()) {
				return +1;
			} else {
				return 0;
			}
		}
	}

}